/* **********************************************************************************
Copyright (c) 2009, MARK OSBORNE [LOUD-FAT-BLOKE.CO.UK] &
                    INTEROUTE COMMUNICATIONS LTD
                    ALL RIGHTS RESERVED.

THIS SOFTWARE IS FOR PRIVATE, NON-COMMERCIAL USE ONLY.
USE WITHIN ANY ORGANIZATION OR FOR COMMERCIAL PURPOSES IS STRICTLY PROHIBITED.
THE FULL LICENSE AGREEMENT IS PROVIDED IN THE FILE NAMED  "LICENSE"
PROVIDED WITH THIS DISTRIBUTION.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA,
OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY
OF SUCH DAMAGE.

****************************************************************************************/

int debug = 0;

/*
 OBESEUSVB       ENVISAGED DESIGNED AND ORIGINALLY CODED BY LOUD-FAT-BLOKE
                 Recoded and extensively extended by Simon Ratcliffe, (at great pains, undergoing extensive therapy).
****************************************************************************
* This program fundamentally captures potential attack traffic and analyses
* it and reports as/if appropriate.
*
* This beta version of the program does the following
*       1 - Locates the busiest subnets. (4 minimum).
*       2 - Identifies the busiest destination ip on that subnet.
*       3 - Identifies the busiest protocol on the ip address established at 2.
*       4 - If the busiest protocol is TCP continue to 5 otherwise go to 1.
*       5 - Identifies the busiest destination port on the ip address and protocol, (2 and 3).
*       6 - Passes profile past Synflood signature and raises an alert as necessary.

* Versions
*       This is a beta version of the code.
*       Future versions of the core code shall include additional signatures, robustness
*       checks, strength assurity and much of the code to be ported to our FPGA
*       accelerated platform.
****************************************************************************
*/
int fd;
#include <fcntl.h>
#include <pcap.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <ctype.h>
#include <errno.h>
#include <signal.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/stat.h>
#include <time.h>
#include <unistd.h>
#include <netinet/in.h>
#include <arpa/inet.h>

#define __FAVOR_BSD
#include <netinet/ip.h>
#include <netinet/udp.h>
#include <netinet/tcp.h>

#include <sys/syslog.h>

/* default snap length (maximum bytes per packet to capture) */
#define SNAP_LEN 1518

/* ethernet headers are always exactly 14 bytes [1] */
#define SIZE_ETHERNET 14

/* Ethernet addresses are 6 bytes */
#define ETHER_ADDR_LEN	6

/* SIGALRM durations */
#define ALARM_DUR1 10
#define ALARM_DUR2 15
#define ALARM_DUR3 180

/* 5% arbitrary capture to allow meaningful analysis */
#define SOME_PACKETS 500

/* Data file log */
#define DATA_FILE "/var/log/obeseus_data.log"

/* SYN flood log */
#define ALERT_FILE "/var/log/obeseus_alert.log"

/* File management */
#define TMPDIR "/tmp"
#define PREFIX "OBESS"

/* Size of blacklist/aux array. */
#define BLK_NUM 2000
#define AUX_BLK_NUM 500

/* Maximum no repeated top subnets, arbitrary ignore value */
#define MAX_NO_24 3
#define IGNORE_24 1000

/* Ip string length */
#define IP_LEN 16

/* No. of consecutive SYNS. */
#define CONSEC_FLAGS 10

/* This defines the list of /24s to ignore for an attempt at ai */
#define BLACKLIST_FILE "/var/log/blacklist"
#define SUBNET_LEN 13
#define BLACK_LEN 21

/* Alert reliability */
#define GOOD_INDICATOR 20
#define STRONG_INDICATOR 30
#define STRONGER_INDICATOR 50
#define VERY_STRONG_INDICATOR 80
#define POSSIBLE 50
#define STRONG 70
#define VERY_STRONG 80

/* Loop delay */
#define FIVE_SECS 5

/* Callback function pointer */
static void (*callback) (u_char*, const struct pcap_pkthdr*, const u_char*) = NULL;

/* Protos */
static int check_blacklist(uint32_t, int);

typedef struct 
{
	char subnet[SUBNET_LEN];
	int count;
} blktype;

typedef struct
{
	uint32_t qual;
	int count;
} qualtype;

/* Blacklist quick access array and stack num. */
static blktype blacklist[BLK_NUM];
static int blk_ptr = 0;

/*Auxiliary blacklist */
static blktype aux_blacklist[AUX_BLK_NUM];
static int aux_blk_ptr = 0;

/* Ethernet header */
struct sniff_ethernet {
        u_char  ether_dhost[ETHER_ADDR_LEN];    /* destination host address */
        u_char  ether_shost[ETHER_ADDR_LEN];    /* source host address */
        u_short ether_type;                     /* IP? ARP? RARP? etc */
};

#define IP_HL(ip)               (((ip)->ip_hl) & 0x0f)
#define IP_V(ip)                (((ip)->ip_hl) >> 4)

static void got_packet(u_char *args, const struct pcap_pkthdr *header, const u_char *packet);
static void count_proto(u_char proto);

 /* print help text
 */
void print_app_usage(void)
{
	printf("Usage: obeseusB <i/f> ip");
	printf("\n");
  	printf("\n");

  	return;
}

char *dev = NULL;			/* pcap name */
#define  STAGE1_CAPTURE_SIZE 10000
#define  STAGE2_CAPTURE_SIZE 10000
#define  STAGE3_CAPTURE_SIZE 10000
#define  STAGE4_CAPTURE_SIZE 10000
#define  STAGE5_CAPTURE_SIZE 10000
#define TABLE_SIZE 10000

/* 1/1000 of the original table size to accomodate the
TOP_PC_DEC  and greater of all recorded address counts. */
#define TOP_PC_SIZE TABLE_SIZE/1000
/* Defines the top % of dest address counts to store.
Currently set at 40%, i.e top 60% are recorded. */ 
#define TOP_PC_DEC 0.4

/* Defines the minimum and reduction factor for count sizes. */
#define MIN_PC 0.1
#define QUAL_FACT 0.9

/* Defines for SYN FLOOD attack */
/* Defines ratio of fins to syns/rsts.
 * Any less than this dictates a likely Synflood
 * Resetflood as the connection is imbalanced. */
#define SYN_FINRATE 0.85
#define TWENTY_PERCENT_DEC 0.2

/* Defines for RST ICMP and UDP attacks */
#define FOURTY_PERCENT_DEC 0.4
#define SIXTY_PERCENT_DEC 0.6
#define EIGHTY_PERCENT_DEC 0.8

/* Probe unique id support */
#define HOSTLEN 20
#define UNQIDLEN 2

/* 12 hours in seconds */
#define TWELVE_HOURS 60 * 60 * 12

static FILE* data_fptr = NULL;
static long tcp_cnt, udp_cnt, icmp_cnt; 
static int syn_cnt, ack_cnt, len_cnt, last_len, fin_cnt = 0;
static int rst_cnt = 0; 
static int rfc1918_cnt = 0;
static int band_ptr = 0;
static int band_top = 0;
static int prev_count = 0;
static size_t prev_time = 0;
static unsigned long prev_size;
static uint32_t table[TABLE_SIZE];
static uint32_t table_syn[TABLE_SIZE];
static signed short type_table[TABLE_SIZE];
static int table_cnt[TABLE_SIZE];
static int table_syn_cnt[TABLE_SIZE];
static int max_syn_cnt[TABLE_SIZE];
static int table_max_syn_cnt[TABLE_SIZE];
static int max_syns = 0;
static int max_packet_cnt = 0;
static int max_num_srcs = 0;
static int max_num_syn_srcs = 0;
static int max_syn_packet_cnt = 0;
static unsigned long total_len = 0;
static unsigned long total_attack_count = 0;
static unsigned long length[TABLE_SIZE];
static unsigned short chksum[TABLE_SIZE];
static unsigned short chksumlen_cnt[TABLE_SIZE];
static unsigned short last_chksum = 0;
static unsigned long max_chksumlen_count = 0;
static float count_rate[MAX_NO_24];
static float size_rate[MAX_NO_24];
static uint32_t top_24_pc[TOP_PC_SIZE];
static int top_24_count = 0;            /* Holds no. subnets which qualify > TOP_PC_DEC */
static int top_24_pc_no = 0;            /* counter to process top % of 24 subnets */
static int top_srcip_count = 0;		/* Number of src addresses for this host */
uint32_t top_srcip_pc[TOP_PC_SIZE];	/* ip addr of the srcs */
qualtype syn_qualifiers[TOP_PC_SIZE];
static int count = 1;                   /* packet counter */
static int stage3_count = 0;            /* stage 3 packet count */
static int max_port_cnt, port_table_cnt[1000], port_table[1000];
static int max_port;
static int timeout = 0;
static int weighting = 0;
struct in_addr print_ip;  /*  address */
struct in_addr max_ip;  /* winner of the most popular dst  address */
static char errbuf[PCAP_ERRBUF_SIZE];          /* error buffer */
static pcap_t *handle;                         /* packet capture handle */
static char filter_exp[120] =  "\0                                                                      ";
char filter_port[30] = {""};
static char filter_proto[4] ={""};
static char max_ip_addr[IP_LEN] = {""};
static struct bpf_program fp;                  /* compiled filter program (expression) */
static bpf_u_int32 mask;                       /* subnet mask */
static bpf_u_int32 net;                        /* ip */
static pcap_dumper_t* dumpd = NULL;
static char tdate_buf[29] = {""};
static char* unqfnam = NULL;
static time_t current;
static time_t aux_blacklist_last = 0;
static int aux_blk_notify = 0;
static int failed_status = 0;
static char current_attack[22] = {""};
static char current_dst_addr[IP_LEN] ={""};
static time_t start_time = 0;
int flag_dec = 0;

/* Simple function to clean up unwated pcap logs. */
static remove_pcap(void)
{
	unlink(unqfnam);
	if (unqfnam != NULL)
		free(unqfnam);
	unqfnam = 0;
}

/* Function to report bandwidth stats. */
static void report_bandwidth(void)
{
	int i;
	long total_count = 0;
	long total_size = 0;
	FILE* alert_fptr = NULL;

	if (band_ptr > 0)
	{
		if (band_ptr == 1)
		{
			total_count = prev_count;
			total_size = prev_size;
		}
		else if (band_ptr > 1)
		{
			if (band_top != 0)
				band_ptr = band_top;

			for (i=1; i<=band_ptr; i++)
			{
				total_count += count_rate[i];
				total_size += size_rate[i];
				count_rate[i] = 0;
				size_rate[i] = 0;
			}
		}

		if (total_count > 0 && total_size > 0 && start_time > 0 && current > start_time)
		{
			alert_fptr = fopen(ALERT_FILE, "a+");

			if (alert_fptr != NULL)
			{
				float minutes = 0;
				float delta_time = (float)(current-start_time);
				minutes = (float)delta_time/60;

				fprintf(alert_fptr, "***%s ATTACK BANDWIDTH STATS--------------------------------------***\n", current_attack);
				if (minutes > 1.0)
					fprintf(alert_fptr, "%s attack complete duration was %3.2f minutes\n", current_attack, minutes);
				else
					fprintf(alert_fptr, "%s attack complete duration was %lu seconds\n", current_attack, current-start_time);
				fprintf(alert_fptr, "%s attack complete average size is %d bytes sec\n", current_attack, (int)total_size/band_ptr);
				fprintf(alert_fptr, "%s attack complete average count is %d packets sec\n", current_attack, (int)total_count/band_ptr);
				fprintf(alert_fptr, "***------------------------------------------------------------------***\n");
                       		fclose(alert_fptr);
			}
		}
		band_ptr = 0;
		band_top = 0;
		prev_time = 0;
		prev_size = 0;
		prev_count = 0;
		start_time = 0;
		total_attack_count = 0;
		total_len = 0;
		memset(current_attack, 0, 22);
		memset(current_dst_addr, 0, IP_LEN);
	}
}

/* Function to determine bandwidth of attack. */
static void calc_bandwidth(char* signature, char* dst_addr, int attack_count)
{
	if (signature != NULL && dst_addr != NULL)
	{
		if ((strstr(current_attack, signature) == NULL) ||
		     (strstr(current_dst_addr, dst_addr) == NULL))
		{
			if (strlen(current_attack) > 0)
				report_bandwidth();
			strcpy(current_attack, signature);
			strcpy(current_dst_addr, dst_addr);
 		}
		if (band_ptr > 0)
		{
			if (band_ptr == MAX_NO_24)
			{
				band_top = band_ptr;
				band_ptr = 0;
			}
			total_attack_count += attack_count;
			count_rate[band_ptr] = (float)((total_attack_count - prev_count)/(current - prev_time));
			size_rate[band_ptr] = (float)((total_len - prev_size)/(current - prev_time));
		}
		if (band_ptr == 0 && band_top == 0)
			start_time = current;

		prev_time = current;
                prev_count += attack_count;
                prev_size = total_len;
		band_ptr++;
	}
	else 
	{
		if (band_ptr > 0 && (strlen(current_attack) > 0))
			report_bandwidth();
		total_len = 0;
		total_attack_count = 0;
	}
}

/* This function initialises the blacklist array with data from the blacklist file.
*/
static void init_blacklist(void)
{
	FILE* blk_fptr = NULL;
	char read_buf[BLACK_LEN] = {""};
	char subnet[SUBNET_LEN] = {""};
	int blk_count = 0;
	int cnt = 0;

	if (access(BLACKLIST_FILE, R_OK) == 0)
	{
		blk_fptr = fopen(BLACKLIST_FILE, "r");

        	if (blk_fptr != NULL)
        	{
			blk_ptr = 0;
			while (blk_count<BLK_NUM)
			{
				memset(blacklist[blk_count].subnet, 0, SUBNET_LEN);
				blacklist[blk_count].count = 0;
				memset(read_buf, 0, BLACK_LEN);

                		if (fgets(read_buf, BLACK_LEN, blk_fptr) != NULL)
                		{
					memset(subnet, 0, SUBNET_LEN);
					cnt = 0;
					sscanf(read_buf, "%s %d", &subnet, &cnt);
					strncpy(blacklist[blk_count].subnet, subnet, SUBNET_LEN);
					blacklist[blk_count].count = cnt;
					blk_ptr++;
				}
				blk_count++;
			}
                	fclose(blk_fptr);
		}
	}
}

/* Update the blacklist file by removing it and generating a new 
 * one with the contents from the blacklist array 
*/
static void update_blacklist_file(void)
{
        FILE* blk_fptr = NULL;
	char writebuf[BLACK_LEN] = {""};
	int i;
	int subnet_size = 0;

	if (blk_ptr > 0)
	{
		/* Remove file if it exists */
        	if (access(BLACKLIST_FILE, R_OK) == 0)
			unlink(BLACKLIST_FILE);

        	blk_fptr = fopen(BLACKLIST_FILE, "w");

        	if (blk_fptr != NULL)
        	{
			for (i=0; i<blk_ptr; i++)
                	{
				/* If seen >= MAX_NO_24 times and not currently an ignore set to ignore */
				if (blacklist[i].count < IGNORE_24 && blacklist[i].count >= MAX_NO_24)
					blacklist[i].count = IGNORE_24;
				subnet_size = strlen(blacklist[i].subnet);
				if (blacklist[i].count > 0 && blacklist[i].subnet[subnet_size - 1] == '.')
                                {
                                        sprintf(writebuf, "%s %d\n", blacklist[i].subnet, blacklist[i].count);
                                        fprintf(blk_fptr, writebuf);
                                        memset(writebuf, 0, BLACK_LEN);
                                }
				subnet_size = 0;
			}
			fclose(blk_fptr);
		}
	}
}

/* Check the time and see if the aux blacklist needs clearing out. */
static void init_aux_blacklist(void)
{
	int i;

	if ((current - TWELVE_HOURS) > aux_blacklist_last)
	{
		for (i=0; i<AUX_BLK_NUM; i++)
		{
			aux_blacklist[i].count = 0;
			memset(aux_blacklist[i].subnet, 0, SUBNET_LEN);
		}
		aux_blk_ptr = 0;
		aux_blacklist_last = current;
		aux_blk_notify = 0;
		aux_blk_notify = 0;
	}
}

/* Trivial function to aid readability in output data file upon a continue condition. */
static void output_seperator(void)
{
        fprintf(data_fptr, "----------------------------------------------------------------------------  \n");
}

/* Static function to check that enough packets have been received for a meaningful analysis */
static int min_threshold_met(int stage)
{
	int threshold_met = 0;
	int local_count = 0;

	if (stage == 3)
		local_count = stage3_count;
	else
		local_count = count;	

	if (local_count < SOME_PACKETS)
	{
		fprintf(data_fptr, "Minimum packet count of %d NOT met for Stage %d processing.\n", SOME_PACKETS, stage);
		output_seperator();
	}	
	else
		threshold_met = 1;

	return threshold_met;	
}
/* Calculates modulo11 checksum of payload */
static unsigned short modulo11(const char* payload, int payload_size)
{
        int i;
        unsigned short chksum = 0;
        unsigned long sum = 0;

	if (payload != NULL && payload_size > 0)
	{
		for (i=0; i<=payload_size; i++)
                	sum += ((int)payload[i] * (i+2));
        	chksum = sum % 11;
	}
        return chksum;
}

/* Displays the address table for debugging purposes only */ 
void display_table()
{
	int n;

  	for (n = 0; n < TABLE_SIZE -1; n++) 
  	{
    		print_ip.s_addr = htonl(table[n]  << 8);
    		printf(" table n %i ",    table[n])  ;
    		printf("    %i      %s : %i \n", n,  inet_ntoa(  print_ip) , table_cnt[n] ) ;
  	}
  	return;
}

/* Static function to pass back the top TOP_PC_SIZE /24s (uint32_t type), 
 * which have the highest qualifying count.
*/
static void generate_top_pc(uint32_t* storage, int* counter, int enough)
{
        int i;
	int y = 0;
	int z;
	int pos;
	int status = 0;
	int table_type = 5;
	float qualify_cnt = 0;

	qualtype qualifiers[TOP_PC_SIZE];

	qualify_cnt = (float)(max_packet_cnt * TOP_PC_DEC);
	
	/* Initialise */
	for (i = 0; i < TOP_PC_SIZE; i++)
	{
		qualifiers[i].qual = 0;
		qualifiers[i].count = 0;
        }

	while (y < enough)
	{
        	/* Find values which qualify */
        	for ( i = 0; i < TABLE_SIZE && y < TOP_PC_SIZE; i++)
        	{
			if ((float)table_cnt[i] >= qualify_cnt)
			{
				status = check_blacklist(table[i], 0);
                		if (status <= 1)
				{
                        		qualifiers[y].qual = table[i];
					qualifiers[y].count = table_cnt[i];
					table_cnt[i] = 0;
					y++;
				}
				if (status == 0)
					check_blacklist(table[i], 1);
			}
        	}
		
		if (y < enough)
		{
			qualify_cnt*=QUAL_FACT;
			if (qualify_cnt <= (float)(max_packet_cnt * MIN_PC))
				break;
		}
	}

	/* Sort and write the qualifiers */
        for ( i = 0; i < y; i++)
        {
		
		int largest = 0;
		for (z = 0; z < y; z++)
        	{
			if (qualifiers[z].count > largest)
			{
				largest = qualifiers[z].count;
				pos = z;
			}

		}
		qualifiers[pos].count = 0;
		storage[(*counter)++] = qualifiers[pos].qual;
        }
}

/* Alarm handler functions.
 * SigAlrmHandler performs cleaning up after a forced pcap loop timeout
 * has occurred. SigAlrmHandlerCount checks the packet count to see if
 * at least SOME_PACKETS have been caught. If not it gives the process
 * another ALARM_DUR1 seconds. SigBucketHandler monitors the bucket
 * processing to exit in a controlled fashion if necessary.
*/
static void SigAlrmHandler(int signal)
{
    pcap_freecode(&fp);
    pcap_close(handle);
    timeout = 1;
    fprintf(data_fptr, "PLEASE NOTE: Capture was cut short due to timeout in PCAP\n");
}

static void SigAlrmHandlerCount(int sig)
{
    if (count > SOME_PACKETS)
        SigAlrmHandler(sig);
    else
    {
        alarm(ALARM_DUR1);
        signal (SIGALRM, SigAlrmHandler);
    }    
}

static void SigBucketHandler(int signal)
{
	pcap_freecode(&fp);
    	pcap_close(handle);
        fprintf(data_fptr, "PLEASE NOTE: There are no /24s that are not in the blacklist..\n");
        output_seperator();
        fflush(data_fptr);
	failed_status = -1;
}

/* Static function to report if the passed address falls within
 * the rfc 1918, (private VPN), range, meaning these are
*  definitely spoof. */
static int is_rfc1918(uint32_t x)
{
/*
10.0.0.0/8
*/
        if ((167772160 < x) && (x < 184549375))
                return 1;
/* 192.168
*/
        if ((3232235520   < x) && (x <  3232301055))
                return 1;
/*

172.32
*/
        if ((2886729728 < x) && (x <  2887843839))
                return 1;
/*
127.0.0.1 & 224.0.0.0

*/
        if (x == 2130706433)
                return 1;

        if (x == 3758096384 )
                return 1;

        if (x == 0)
                return 1;

        return 0;
}

/* Store the address counts. */
static void add_table(uint32_t passed_addr)
{
        int n;
        print_ip.s_addr =  passed_addr;

        for ( n = 0; n < TABLE_SIZE ; n++)
        {
                if ( table[n] == 0 )
                {
                        table[n] =  print_ip.s_addr;
                        table_cnt[n] = 1;
                        return;
                }
                if (table[n] == print_ip.s_addr)
                {
                        table_cnt[n]++;
                        if (table_cnt[n] >  max_packet_cnt)
                        {
                                max_ip.s_addr =  print_ip.s_addr;
                                max_packet_cnt = table_cnt[n];
                        }
                        return;
                }

        }
        return;
}

/* Store the address counts AND takes a u_char parameter to count sequential SYNs
 * This function is also used for counting src addresses for non SYN attacks, here
 * the SYN flag is set in the call to add_table_count */
static void add_table_count(uint32_t pass_addr, u_char th_flags)
{
        int n;
	int state = 4;
	struct in_addr pass_ip; 
        pass_ip.s_addr = pass_addr;

	if (pass_addr == 0)
		return;

	if (th_flags & TH_SYN)
		state = 6;

	/* Perform source address checking for SYN flags. */
        for (n = 0; n < TABLE_SIZE; n++)
        {
                if (table[n] == 0)
                {
			table_cnt[n] = 1;
			max_num_srcs++;
			table[n] = pass_ip.s_addr;
			if (state == 6)
			{
				table_syn_cnt[n] = 1;
				max_syn_cnt[n] = table_syn_cnt[n];
				type_table[n] = state;
			}
			else
				type_table[n] = 4;
                        return;
                }
                if (table[n] == pass_ip.s_addr)
                {
			table_cnt[n]++;
			if (state == 4)
			{
				table_syn_cnt[n] = 0;
				type_table[n] = 4;
			}
			else if (type_table[n] != state)
			{
				if (state == 6)
				{
					table_syn_cnt[n] = 1;
					type_table[n] = 6;
				}
                        }
                        else if (state == 6)
			{
				table_syn_cnt[n]++;
				if (table_syn_cnt[n] > max_syn_packet_cnt)
					max_syn_packet_cnt = table_syn_cnt[n];
				if (table_syn_cnt[n] >  max_syn_cnt[n])
				{
					max_syn_cnt[n] = table_syn_cnt[n];
					if (max_syn_cnt[n] == CONSEC_FLAGS)
						max_num_syn_srcs++;
				}
			}
                        return;
		}
        }
	return;
}

/* Stores the payload length and checksum counts. */
static void add_data_count(unsigned short checksum, int len)
{
        int n;

        for (n = 0; n < TABLE_SIZE; n++)
        {
                if (length[n] == 0)
                {
			length[n] = len;
			chksum[n] = checksum;
			chksumlen_cnt[n] = 1;
			return;
		}
                if (length[n] == len &&
			chksum[n] == checksum)
		{
			chksumlen_cnt[n]++;
			if (chksumlen_cnt[n] > max_chksumlen_count)
				max_chksumlen_count = chksumlen_cnt[n];
		}
        }
        return;
}

/* Store the port count. */
static void add_port_table( int passed_port )
{
	int n;
          
	for ( n = 0 ; n < 1000 -1 ; n++) 
     	{
     		if ( port_table[n] == 0 )
        	{
         		port_table[n] =  passed_port;
         		port_table_cnt[n] = 1;
         		return;
         	}
     		if (port_table[n] == passed_port)
        	{
         		port_table_cnt[n]++;
         		if ( port_table_cnt[n] >  max_port_cnt)
           		{
           			max_port =  passed_port;
           			max_port_cnt = port_table_cnt[n]; 
            		}
         		return ;
         	} 
      	} 
	return;
}

/* Log the protocol types. */
static void count_proto(u_char proto)
{
        /* determine protocol */
        switch(proto )
        {
                case IPPROTO_TCP:
                        tcp_cnt++;
                        break;
                case IPPROTO_UDP:
                        udp_cnt++;
                        break;
                case IPPROTO_ICMP:
                        icmp_cnt++;
                        break;
                default:
                        break;
        }

        if ( debug == 1)
        {
                printf(" Max packet count  tcp %u udp %u icmp %u  \n", tcp_cnt, udp_cnt,icmp_cnt );
        }

        return;
}

/* Adds/checks the blacklist. */
static int check_blacklist(uint32_t passed_sub, int modify)
{
        int i = 0;
	int size24 = 0;
	struct in_addr passed_subnet;
	char subnet_24[SUBNET_LEN] = {""};

	passed_subnet.s_addr = htonl(passed_sub << 8);
	memset(subnet_24, 0, SUBNET_LEN);
	sprintf(subnet_24, inet_ntoa(passed_subnet));

	/* Cut off last octet to make /24 */
	for(i=strlen(subnet_24); i>0; i--)
	{
		if (subnet_24[i] == '.')
		{
			size24 = ++i;
			break;
		}
	}
	
	if (modify == 0)
	{
		/* Check/update the blacklist */
        	for (i=0; i<blk_ptr; i++)
        	{
                	if (strncmp(blacklist[i].subnet, subnet_24, size24) == 0)
			{
				if (blacklist[i].count < MAX_NO_24)
				{
					blacklist[i].count++;
					return 1;
				}
				if (blacklist[i].count >= MAX_NO_24)
					return 2;
                	}
        	}

		/* Check/update the auxiliary blacklist */
		if (blk_ptr == BLK_NUM)
		{
			if (aux_blk_ptr == AUX_BLK_NUM)
			{
				if (aux_blk_notify < 3)
				{
					fprintf(data_fptr, "The auxilliary blacklist has reached its capacity of %d. Ignoring.\n", AUX_BLK_NUM);
					aux_blk_notify++;
				}
				return 1;
			}
			for (i=0; i<aux_blk_ptr; i++)
        		{
                		if (strncmp(aux_blacklist[i].subnet, subnet_24, size24) == 0)
                		{
                        		if (aux_blacklist[i].count < MAX_NO_24)
					{
						aux_blacklist[i].count++;
						return 1;
					}
                        		if (aux_blacklist[i].count >= MAX_NO_24)
						return 2;
				}
                	}
        	}
		return 0;
	}
	else
        {      
		if (BLK_NUM > blk_ptr)
		{
			memset(blacklist[blk_ptr].subnet, 0, SUBNET_LEN);
			strncpy(blacklist[blk_ptr].subnet, subnet_24, size24);
			blacklist[blk_ptr].count = 1;
			blk_ptr++;
        	}
		else if (blk_ptr == BLK_NUM)
		{
			fprintf(data_fptr, "The blacklist has reached its capacity of %d. Perhaps increase\n", BLK_NUM);
        		fprintf(data_fptr, "its size or review contents of %s.\n", BLACKLIST_FILE);
			
			if (aux_blk_ptr < AUX_BLK_NUM)
			{
				fprintf(data_fptr, "This address shall be written to the temporary blacklist which shall\n");
				fprintf(data_fptr, "be cleared out after 24 hours\n");
				memset(aux_blacklist[aux_blk_ptr].subnet, 0, SUBNET_LEN);
				strncpy(aux_blacklist[aux_blk_ptr].subnet, subnet_24, size24);
				aux_blacklist[aux_blk_ptr].count = 1;
				aux_blk_ptr++;
			}
		}
		return 0;
	}
	return 2;
}

/* Store the bucket count. */
void get_bucket(u_char *args, const struct pcap_pkthdr *header, const u_char *packet)
{
	/* declare pointers to packet headers */
	const struct sniff_ethernet *ethernet;  /* The ethernet header [1] */
	const struct ip *ip;        	        /* The IP header */

	int size_ip;
	int n;
 	uint32_t temp_addr;
	
	count++;

	/* define ethernet header */
	ethernet = (struct sniff_ethernet*)(packet);
	
	/* define/compute ip header offset */
	ip = (struct ip*)(packet + SIZE_ETHERNET);
	size_ip = IP_HL(ip)*4;
	if (size_ip < 20) 
	{
		fprintf(data_fptr, "Invalid IP header length: %u bytes\n", size_ip);
		return;
	}

	/* print source and destination IP addresses 
	*/
	if ( debug == 1)
	{
		printf("         To: %s\n", inet_ntoa(ip->ip_dst));
		printf("         frx %s\n", inet_ntoa(ip->ip_src));
	}
	fflush(0);

	temp_addr = ntohl(ip->ip_dst.s_addr  ) ;
	temp_addr = temp_addr >> 8;
	add_table(temp_addr);
	return;
}

/* Get the address of the attack. */
static void got_packet(u_char *args, const struct pcap_pkthdr *header, const u_char *packet)
{
	
	/* declare pointers to packet headers */
	const struct sniff_ethernet *ethernet;  /* The ethernet header [1] */
	const struct ip *ip;                    /* The IP header */
	const struct tcphdr *tcp;               /* The TCP header */
	const char *payload;                    /* Packet payload */

	int size_ip;
	int size_tcp;
	int size_payload;
	int size_out;
	int n;
 	uint32_t temp_addr;
	
	count++;
	
	/* define ethernet header */
	ethernet = (struct sniff_ethernet*)(packet);
	
	/* define/compute ip header offset */
	ip = (struct ip*)(packet + SIZE_ETHERNET);
	size_ip = IP_HL(ip)*4;
	if (size_ip < 20)
	{
		fprintf(data_fptr, "Invalid IP header length: %u bytes\n", size_ip);
		return;
	}

	/* print source and destination IP addresses 
	*/
	if (debug == 1)
	{
		printf("         To: %s\n", inet_ntoa(ip->ip_dst));
	}
 	temp_addr = ntohl(ip->ip_dst.s_addr); 

	add_table(temp_addr);
        return;
}

/* Dissect the ip attack. */
void get_ip_attack(u_char *args, const struct pcap_pkthdr *header, const u_char *packet)
{	
	/* declare pointers to packet headers */
	const struct sniff_ethernet *ethernet;  /* The ethernet header [1] */
	const struct ip *ip;                    /* The IP header */
	const struct tcphdr *tcp;               /* The TCP header */
        const struct udphdr *udp;               /* The udp  header */

	const char *payload;                    /* Packet payload */

	int size_ip;
	int size_tcp;
	int size_payload;
	int n;
 	uint32_t temp_addr;

	stage3_count++;
	
	/* define ethernet header */
	ethernet = (struct sniff_ethernet*)(packet);
	
	/* define/compute ip header offset */
	ip = (struct ip*)(packet + SIZE_ETHERNET);
	size_ip = IP_HL(ip)*4;
	if (size_ip < 20) 
	{
		if (debug == 1)
			printf("   * Invalid IP header length: %u bytes\n", size_ip);
		return;
	}

	/* Total length for bandwidth calcs. */
	total_len = ip->ip_len;

	/* Print valid ip packets to temp log file */
        pcap_dump((u_char *)dumpd, header, packet);

	/* determine protocol */	
	count_proto(ip->ip_p) ;

	return;
}

/* Dissect the tcp attack. */
void get_tcp_attack(u_char *args, const struct pcap_pkthdr *header, const u_char *packet)
{
	/* declare pointers to packet headers */
	const struct sniff_ethernet *ethernet;  /* The ethernet header [1] */
	const struct ip *ip;                    /* The IP header */
	const struct tcphdr *tcp;               /* The TCP header */
	const char *payload;                    /* Packet payload */
	int size_payload;
	int size_ip;
	int size_tcp;
	int n;
 	uint32_t temp_addr;
	uint32_t src_addr;
	
	count++;
	
	/* define ethernet header */
	ethernet = (struct sniff_ethernet*)(packet);
	
	/* define/compute ip header offset */
	ip = (struct ip*)(packet + SIZE_ETHERNET);
	size_ip = IP_HL(ip)*4;
	if (size_ip < 20) {
		printf("   * Invalid IP header length: %u bytes\n", size_ip);
		return;
	}

	/* determine protocol */	
	switch(ip->ip_p) 
	{
		case IPPROTO_TCP:
			break;
		case IPPROTO_UDP:
                        if (debug == 1)
				printf("   Protocol: UDP - but i am in get_tcp_attack\n");
                        return;
	
		default:
			if (debug == 1)
			{
				printf("   Protocol: unknown-skipping");
				printf("    - but i am in get_tcp_attack\n");
			}
			return;
	}
	
	/*
	 *  OK, this packet is TCP.
	 */
	
	/* define/compute tcp header offset */
	tcp = (struct tcphdr*)(packet + SIZE_ETHERNET + size_ip);
	size_tcp = tcp->th_off*4;
	
	if (size_tcp < 20) 
	{
		printf("   * Invalid TCP header length: %u bytes\n", size_tcp);
		return;
	}
        if ( debug == 1 )
        {
        	/* print source and destination IP addresses */
		printf("       From: %s\n", inet_ntoa(ip->ip_src));
		printf("         To: %s\n", inet_ntoa(ip->ip_dst));
        	/* print source and destination ports         
	 	* */
        	printf("sport dport %i %i ",
                        ntohs(tcp->th_sport)  ,
	                ntohs(tcp->th_dport));
        }

	/* Print valid tcp packets to temp log file */
	pcap_dump((u_char *)dumpd, header, packet);
	
	/* define/compute tcp payload (segment) offset */
	payload = (u_char *)(packet + SIZE_ETHERNET + size_ip + size_tcp);

	/* compute tcp payload (segment) size */
	size_payload = ntohs(ip->ip_len) - (size_ip + size_tcp);
	
	if (size_payload > 0)
	{
		/* Store payload sizes and checksums */
		add_data_count(modulo11(payload, size_payload), size_payload);
	}

	/*    detect a syn flood */
	if (tcp->th_flags & TH_SYN)
		syn_cnt++;
	/*    detect a rst  flood */
	if (tcp->th_flags & TH_RST)
    		rst_cnt++;
        /*    count ack packets */ 
        if (tcp->th_flags & TH_ACK)
                ack_cnt++;
        if (tcp->th_flags & TH_FIN)
                fin_cnt++;

	/* Update the src address list for this capture. */
	add_table_count(ip->ip_src.s_addr, tcp->th_flags);
	
	/* look for spoof src add */
	src_addr = ntohl(ip->ip_src.s_addr);
	rfc1918_cnt = rfc1918_cnt + is_rfc1918(src_addr);

	return;
}

/* Determine port count. */
void get_port_attack(u_char *args, const struct pcap_pkthdr *header, const u_char *packet)
{	
	/* declare pointers to packet headers */
	const struct sniff_ethernet *ethernet;  /* The ethernet header [1] */
	const struct ip *ip;                    /* The IP header */
	const struct tcphdr *tcp;               /* The TCP header */
        const struct udphdr *udp;               /* The udp  header */

	const char *payload;                    /* Packet payload */


	int size_ip;
	int size_tcp;
	int size_payload;
	int size_out;
	int n;
 	uint32_t temp_addr;
		
	count++;

	
	/* define ethernet header */
	ethernet = (struct sniff_ethernet*)(packet);
	
	/* define/compute ip header offset */
	ip = (struct ip*)(packet + SIZE_ETHERNET);
	size_ip = IP_HL(ip)*4;
	if (size_ip < 20) 
	{
		printf("   * Invalid IP header length: %u bytes\n", size_ip);
		return;
	}
	/* timestamp bollock */
	switch(ip->ip_p) {
		case IPPROTO_TCP:
		case IPPROTO_UDP:
			break;
	
		case IPPROTO_ICMP:
			if (debug == 1)
			{
				printf("   Protocol: ICMP\n");
				printf("   should not get here \n");
			}
                        return; 
	
		default:
			if (debug == 1)
			{
				printf("   Protocol: unknown-skipping\n");
				printf("   should not get here \n");
			}
                        return; 
	}
	
	/*
	 *  OK, this  packet is TCPi or udp - assume udp .
	 */
	
	/* define/compute tcp header offset */
	udp  = (struct udphdr*)(packet + SIZE_ETHERNET + size_ip);
        /* print source and destination IP addresses 
	 * */

	if (debug == 1 )
        	printf(" %i %i ", ntohs(udp->uh_sport), ntohs(udp->uh_dport));
	/* find the most common port              */
	add_port_table(ntohs(udp->uh_dport));
 
	return;
}

/* Static function to build and compile the filter */
static int pcap_build_filter(void)
{
	/*compile the filter expression */
	if (pcap_compile(handle, &fp, filter_exp, 0, net) == -1) {
		fprintf(data_fptr, "Couldn't parse filter %s: %s..\n",
		    filter_exp, pcap_geterr(handle));
		return -1;
	}

	/* apply the compiled filter */
	if (pcap_setfilter(handle, &fp) == -1) 
	{
		fprintf(data_fptr, "Couldn't install filter %s: %s..\n", filter_exp, pcap_geterr(handle));
		return -1;
	}
	return 0;
}

/* Static function to add more telementry assertion nto the alert process. */
static char* report_strength(int weighting)
{
        if (weighting < GOOD_INDICATOR)
                return "";
        else if (weighting >= GOOD_INDICATOR && weighting < POSSIBLE)
                return "POSSIBLE";
        else if (weighting >= POSSIBLE && weighting < STRONG)
                return "STRONG_POSSIBLE";
        else if (weighting >= STRONG && weighting < VERY_STRONG)
                return "STRONG";
        else if (weighting >= VERY_STRONG)
                return "VERY_STRONG";
        else
                return "";
}

/* Static function to pass back the top < TOP_PC_SIZE SYN src addresses. */
static int generate_top_syn_pc(qualtype* qualifiers, int enough)
{
        int i;
	int y = 0;
	int z;
	int pos;
	int status = 0;
	float qualify_cnt = 0;

	for (i = 0; i < TOP_PC_SIZE; i++)
	{
		qualifiers[i].qual = 0;
		qualifiers[i].count = 0;
        }

        /* Find values which qualify */
        for ( i = 0; i < max_syns && y < TOP_PC_SIZE; i++)
        {
		if (max_syn_cnt[i] >= CONSEC_FLAGS)
		{
			qualifiers[y].qual = table_syn[i];
			qualifiers[y].count = max_syn_cnt[i];
			y++;
		}
	}
	/* If not met, return -1. */
	if (y == enough)
		return 0;
	else
		return -1;
}

/* Checks to see if >1 CONSEC_FLAGS SYN src address are available. */
static int check_syn_src_addr(qualtype* syn_qualifiers)
{
	int syn_top_pc = -1;
	int swapped = 1;
	int i = 0;
	int temp_cnt = 0;
	uint32_t temp_addr = 0;
	
	/* Only process if top address SYN count determines it. */
	if (max_syn_packet_cnt >= CONSEC_FLAGS)
	{
		/* Firstly, extract the SYNs into a table. */
		for (i=0; i<TABLE_SIZE; i++)
		{
			if (table[i] != 0 && type_table[i] & 0x02)
			{
				table_syn[max_syns] = table[i];
				table_max_syn_cnt[max_syns] = max_syn_cnt[i];
				max_syns++;
			}
		}
	}

	max_syns--;

	/* Now, sort the max_syn_count table. */
	if (max_syns > 1)
	{
		while (swapped)
		{
			swapped = 0;
			for (i=max_syns; i>0; i--)
			{
        	                if (table_max_syn_cnt[i-1] < table_max_syn_cnt[i])
                	        {
					temp_cnt = table_max_syn_cnt[i-1];
					temp_addr = table_syn[i-1];
					table_max_syn_cnt[i-1] = table_max_syn_cnt[i];
					table_syn[i-1] = table_syn[i];
                                	table_syn[i] = temp_addr;
					table_max_syn_cnt[i] = temp_cnt;
                                	swapped = 1;
					temp_cnt = 0;
					temp_addr = 0;
				}
				
                        }
		}
	}

	/* Now go through the SYNS. */
	if (max_syns > 1)
	{
		while(flag_dec < TOP_PC_SIZE-1)
		{
			syn_top_pc = generate_top_syn_pc(syn_qualifiers, TOP_PC_SIZE-flag_dec);
			if (syn_top_pc == 0)
				return 1;
			flag_dec++;
		}
	}
	return 0;
}

/* Static function to check for a SYN flood attack */
static int is_synflood(void)
{
	static FILE* alert_fptr = NULL;
	char syn_report[190] = {""};
	char src_addrs[260] = {""};
	char attacked[23] = {""};
	char signature[22] = {""};
	float synrate = 0;
        float finrate = 0;
	float syn_finrate = 0;
	float rfc1918_rate = 0;
	int i;
	int synflood = 0;

	struct in_addr src_ip;

	/* Protect from div by 0 */
        if (count == 0 )
                return;

	/* Only perform processing if > 1 src addr with > CONSEC_FLAGS consec syns. */
	if (max_num_syn_srcs > 1)
	{
		synrate = (float)  syn_cnt / count;
		finrate = (float)  fin_cnt / count;
		rfc1918_rate = (float)  rfc1918_cnt / count;	

		if (synrate > 0.0 && finrate > 0.0)
			syn_finrate = (float) finrate / synrate;

		/* Rule 1 */
      		if (synrate > TWENTY_PERCENT_DEC)
		{
			weighting += STRONG_INDICATOR;
			/* Rule 2 */
			if (syn_finrate < SYN_FINRATE)
				weighting += STRONGER_INDICATOR;
		}
		/* Rule 3 */
		if (rfc1918_rate > EIGHTY_PERCENT_DEC)
                	weighting += GOOD_INDICATOR;

		/* Strength Qualifier */
		if (weighting >= STRONG_INDICATOR)
		{
			/* Reset the syn src qualifiers */
			memset(syn_qualifiers, 0, sizeof(qualtype)*TOP_PC_SIZE);

			/* Check for > 1 syn src address. */
			if (check_syn_src_addr(syn_qualifiers) == 1)
			{
				sprintf(syn_report, "%s SYNFLOOD ATTACK DETECTED - status done - id %lu - Target Host %s Target Proto: %s Port: %d", report_strength(weighting), current, max_ip_addr, filter_proto ,max_port );
				fprintf(data_fptr, syn_report);
				fprintf(data_fptr, "\nNo of packets with SYN flag set is %d out of %d captured\n", syn_cnt, count);
				fprintf(data_fptr, "This equates to a SYN packet percentage of %d%%\n", (int)(synrate * 100));
				fprintf(data_fptr, "Percentage of SYNs to FINs %d%%\n", (int)(syn_finrate * 100));
				fprintf(data_fptr, "RFC1918 spoof packet percentage of %d%%\n", (int)(rfc1918_rate * 100));

				/* Concat into a string. */
				sprintf(src_addrs, "SYNFLOOD TOP IPS - status done - id %lu - %d top SYN src addresses ", current, TOP_PC_SIZE-flag_dec);
				/* Update the message. */
				for (i=0; i<TOP_PC_SIZE-flag_dec; i++)
				{
					src_ip.s_addr = syn_qualifiers[i].qual;
					strcat(src_addrs, inet_ntoa(src_ip));
					strcat(src_addrs, " ");
				}
				sprintf(attacked, "attacking Port: %d\n", max_port);
				strcat(src_addrs, attacked);

				openlog("obeseus", LOG_ODELAY, LOG_LOCAL0);
				syslog(LOG_ALERT, syn_report);
				syslog(LOG_ALERT, src_addrs);
				closelog();

				/* Open the alert log file */
				alert_fptr = fopen(ALERT_FILE, "a+");

				if (alert_fptr != NULL)
				{
					strcat(syn_report, "\n");
					fprintf(alert_fptr, syn_report);
					fprintf(alert_fptr, src_addrs);
					fprintf(alert_fptr, "PCAP DUMP FOR THIS ALERT IS: %s\n", unqfnam);
					fclose(alert_fptr);
				}
				synflood = 1;
				sprintf(signature, "SYN %d", current);
				calc_bandwidth(signature, max_ip_addr, syn_cnt);
			}
		}
	}
	weighting = 0;
	return synflood;
}

/* Simple static function to reset re-used data */
void reset_counts(void)
{
	/* Memsets */
	memset(table_cnt, 0, TABLE_SIZE);
	memset(max_syn_cnt, 0, TABLE_SIZE);
	memset(type_table, 0, TABLE_SIZE);
        memset(table, 0, TABLE_SIZE);
	memset(table_syn, 0, TABLE_SIZE);
	memset(table_syn_cnt, 0, TABLE_SIZE);
 	memset(chksum, 0, TABLE_SIZE);
	memset(chksumlen_cnt, 0, TABLE_SIZE);
        memset(length, 0, TABLE_SIZE);
  	memset(port_table_cnt, 0, 1000);
        memset(port_table, 0, 1000);              
	memset(max_ip_addr, 0, IP_LEN);
	memset(filter_exp, 0, 120);
	memset(top_srcip_pc, 0, TOP_PC_SIZE);

	/* Counters */
        count = 0;
        max_packet_cnt = 0;
	max_num_srcs = 0;
	max_num_syn_srcs = 0;
	max_syn_packet_cnt = 0;
	max_syns = 0;
        max_port_cnt = 0;
	rfc1918_cnt = 0;
        tcp_cnt = 0;
	udp_cnt = 0;
	icmp_cnt = 0;
        syn_cnt = 0;
        ack_cnt = 0;
	max_chksumlen_count = 0;
	last_chksum = 0;
	top_srcip_count = 0;
	weighting = 0;
	flag_dec = 0;

	/* Ptr to func */
	callback = NULL;
}

/* Process the top buckets */
int process_buckets(void)
{

	alarm(ALARM_DUR3);
	signal (SIGALRM, SigBucketHandler);

	/* If the bucket count timed out with no buckets, clean up and try again. */
	if (failed_status == -1)
	{
		failed_status = 0;
		alarm(0);
		return -1;
	}

	pcap_loop(handle, (int) STAGE1_CAPTURE_SIZE, get_bucket, NULL);

	memset(top_24_pc,0,TOP_PC_SIZE);

        /* Sort the full list of subnets from the bucket into a top
        % of TOP_PC_SIZE buffer based on count popularity. */
        generate_top_pc(top_24_pc, &top_24_count, 4);
        alarm(0);

        /* cleanup */
        pcap_freecode(&fp);
        pcap_close(handle);
	
	return 0;
}

/* This function processed the packets and hence finds the top talking ip
 * for the specific bucket.
*/
int process_packets(void)
{
	reset_counts();
	strcpy(filter_proto, "");
	strcpy(filter_port, "");

	fprintf(data_fptr, "Stage 2 Find top talking host ip for this bucket..........\n") ;
	/*  build a BPF */
	print_ip.s_addr = htonl( top_24_pc[top_24_pc_no] << 8);
	/* For security */
	if (print_ip.s_addr == 0)
		return -1;
	sprintf(filter_exp ,"dst net %s/24 ", inet_ntoa(print_ip));
	fprintf(data_fptr, "DST NET: Apply filter :%s  \n",  filter_exp);

        handle = pcap_open_live(dev, SNAP_LEN, 1, 10, errbuf);
        if (handle == NULL)
        {
        	fprintf(data_fptr, "Couldn't open file, %s: %s, continuing..\n", dev, errbuf);
                return -1;
        }

        /* Compile the filter expression */
        if (pcap_build_filter() == -1)
		return -1;

        /* Set an alarm for ALARM_DUR2 seconds and setup
        * the handler, (to prevent hanging issues) */
        alarm(ALARM_DUR2);
        signal (SIGALRM, SigAlrmHandlerCount);

	/* Now we can set our callback function. */
	pcap_loop(handle, (int) STAGE2_CAPTURE_SIZE,  got_packet, NULL);

	if (timeout == 0)
	{
		/* cleanup */
		alarm(0);
		pcap_freecode(&fp);
		pcap_close(handle);
        }
        else
		timeout = 0;
	
	print_ip.s_addr = htonl( max_ip.s_addr );
	return 0;
}

/* Main thread */
int main(int argc, char **argv)
{

	int n = 0;
        int no_data_file = 0;
	
	/* check for capture device name on command-line */
	if (argc == 2) 
    		dev = argv[1];
  	else if (argc == 3) 
    		dev = argv[1];
  	else
  	{
  		fprintf(stderr, "error: unrecognized command-line options\n\n");
    		print_app_usage();
    		exit(EXIT_FAILURE);
 	}

	/* Open a data log file */
	data_fptr = fopen(DATA_FILE, "a+");

	if (data_fptr == NULL)
	{
		data_fptr = stdout;
		no_data_file = 1;
	}

	/* Main loop */
	while (1)
        {
		/* Get the current timestamp for attack identification */
        	time(&current);
        	sprintf(tdate_buf, "%s", ctime(&current));

		/* Check if management script has cleared out data file */
		if (access(DATA_FILE, W_OK) != 0)
        		data_fptr = fopen(DATA_FILE, "a+");

		handle = pcap_open_live(dev, SNAP_LEN, 1, 100, errbuf);

		if (handle == NULL)
		{
			fprintf(data_fptr, "Couldn't open device %s: %s EXITING...\n", dev, errbuf);
			sleep(FIVE_SECS);
			continue;
		}

		strcpy(filter_exp, argv[2]);
        	if (pcap_build_filter() == -1)
		{
			fprintf(data_fptr, "Catastrophic: Can't use this filter.. must exit\n");
			exit(EXIT_FAILURE);
		}

		/* Initialise the blacklist */ 
		init_blacklist();

		/* Check auxiliary blacklist age */
		init_aux_blacklist();

        	fprintf(data_fptr, "--------------------------**OBESSIUS STARTED**------------------------------  \n");
        	fprintf(data_fptr, "-------------------------**START OF ITERATION**-----------------------------  \n");
        	fprintf(data_fptr, "TIMESTAMP FOR THIS ITERATION is %s", tdate_buf);
        	fprintf(data_fptr, "----------------------------------------------------------------------------  \n");
        	fprintf(data_fptr, "Stage 1 Bucket Processing..........\n") ;
  		fprintf(data_fptr, "Find the top talking /24\n") ;
  		if (process_buckets() == -1)
			continue;
  	
		fprintf(data_fptr, "There are %d /24 in the bucket which qualify for the top %d per cent count\n", top_24_count, (int)(100 *TOP_PC_DEC)); 
  		fprintf(data_fptr, "Capture complete. read %i \n", count  )  ;                   /* packet counter */
  		fprintf(data_fptr, "----------------------------------------------------------------------------  \n") ;

        	fprintf(data_fptr, "Processing data from top buckets\n") ;
        	fprintf(data_fptr, "----------------------------------------------------------------------------  \n") ;
	
		/* This processing from this point forward is looped to manage process top_pc_count
		number of addresses. */
		for (top_24_pc_no = 0; top_24_pc_no < top_24_count; top_24_pc_no++)
		{
			/* Specific resets */
			max_port = 0;
			stage3_count = 0;
			time(&current);

			if (process_packets() != 0)
			{
				output_seperator();
				continue;
			}
			else if (min_threshold_met(2) == 0)
				continue;
	
                	fprintf(data_fptr, "Top talking ip from /24 bucket %d of %d: Max packet count %s : %i \n", top_24_pc_no + 1, top_24_count,
                	inet_ntoa(print_ip) , max_packet_cnt );

			reset_counts();
                	fprintf(data_fptr, "Stage 3 Find top protocol for this host in this bucket..........\n") ;

                	sprintf( filter_exp ,"dst host  %s ", inet_ntoa(  print_ip) );
			fprintf(data_fptr, "DST HOST: Apply filter  :%s  \n",  filter_exp  );

                	/* open capture device */
                	handle = pcap_open_live(dev, SNAP_LEN, 1, 10, errbuf);
                	if (handle == NULL)
                	{
                		fprintf(stderr, "Couldn't open file, %s: %s, continuing..\n", dev, errbuf);
                		output_seperator();
				continue;
        		}

                	/* Compile the filter expression */
                	if (pcap_build_filter() == -1)
			{
				output_seperator();
				continue;
			}

                	/* Set an alarm for ALARM_DUR2 seconds and setup
                	 * the handler, (to prevent hanging issues) */
                	alarm(ALARM_DUR2);
                	signal (SIGALRM, SigAlrmHandlerCount);

			/* Open unique dumpfile */
			unqfnam = tempnam(TMPDIR, PREFIX);

			/* Capture stage3 in case of pingflood. */
			if (handle != NULL && unqfnam != NULL)
				dumpd = pcap_dump_open(handle, unqfnam);

            		/* Now we can set our callback function
        		*/
           		pcap_loop(handle, (int) STAGE3_CAPTURE_SIZE,  get_ip_attack, NULL);

			if (timeout == 0)
                	{
                    		/* cleanup */
                    		alarm(0);
                    		pcap_freecode(&fp);
                    		pcap_close(handle);
                	}
                	else
                    		timeout = 0;
        
			/* Close dump file. */
                	if (dumpd != NULL)
                        	pcap_dump_close(dumpd);

			/* Check enough packets received. */
        	        if (min_threshold_met(3) == 0)
                	{
				remove_pcap();
                        	continue;
                	}

			/* Report highest protocol count for this iteration */
                	if ((tcp_cnt > icmp_cnt) &&  (tcp_cnt > udp_cnt))
			{
                		fprintf(data_fptr, "Top protocol from /24 bucket %d of %d is tcp\n", top_24_pc_no + 1, top_24_count);
				strcpy(filter_proto,"TCP");
				remove_pcap();
			}
			else
			{
				fprintf(data_fptr, "***This version of Obeseus handles only TCP, continuing..***\n");
				remove_pcap();
				output_seperator();
				continue;
			}
		
                	reset_counts();

			/* Branch for various protocols.
			* This following is for TCP and UDP protocols.
			*/
                	fprintf(data_fptr, "Stage 4 Find top port for this host in this bucket..........\n") ;
                	
			sprintf(filter_exp ,"tcp and dst host %s ", inet_ntoa(print_ip));

			/* open capture device */
	        	handle = pcap_open_live(dev, SNAP_LEN, 1, 10, errbuf);
                	if (handle == NULL)
                	{
				fprintf(stderr, "Couldn't open file, %s: %s, continuing..\n", dev, errbuf);
				output_seperator();
				continue;
			}

			/* compile the filter expression */
			if (pcap_build_filter() == -1)
			{
				output_seperator();
				continue;
			}

			/* Set an alarm for ALARM_DUR2 seconds and setup 
			* the handler, (to prevent hanging issues) */
			alarm(ALARM_DUR2);
			signal (SIGALRM, SigAlrmHandler);

			pcap_loop(handle, (int) STAGE4_CAPTURE_SIZE, get_port_attack, NULL);

			if (timeout == 0)
        	        {
				/* cleanup */
				alarm(0);
				pcap_freecode(&fp);
				pcap_close(handle);
                	}
                	else
				timeout = 0;

			/* Check count. */
        	        if (min_threshold_met(4) == 0)
                	        continue;

        		sprintf(filter_port ," and dst port %i ", max_port);

			fprintf(data_fptr, "Top port from /24 bucket %d of %d is port no.%d with a count of %d\n", top_24_pc_no + 1, top_24_count, max_port, max_port_cnt);

			reset_counts();

			fprintf(data_fptr, "Stage 5 Analyse this attack..........\n");
	
			sprintf( filter_exp ,"tcp and dst host %s  %s", inet_ntoa(print_ip), filter_port);
			callback = get_tcp_attack;

			fprintf(data_fptr, "DST HOST & PORT Apply filter  :%s  \n",  filter_exp);

			/* Write print_ip into max_ip_addr for further use in 
		 	* this iteration as TCP processing reuses the table */
			sprintf(max_ip_addr, "%s", inet_ntoa(print_ip));
		
			/* open capture device */
                	handle = pcap_open_live(dev, SNAP_LEN, 1, 10, errbuf);
                	if (handle == NULL)
                	{
                        	fprintf(stderr, "Couldn't open file, %s: %s, continuing..\n", dev, errbuf);
                        	output_seperator();
				continue;
                	}

			/* compile the filter expression */
                	if (pcap_build_filter() == -1)
			{
				output_seperator();
				continue;
			}

			/* Set an alarm for ALARM_DUR2 seconds and setup
                 	* the handler, (to prevent hanging issues) */
                	alarm(ALARM_DUR2);
                	signal (SIGALRM, SigAlrmHandler);

			/* Open and write packets to a temporary log file if TCP */
			unqfnam = tempnam(TMPDIR, PREFIX);
			if (handle != NULL && unqfnam != NULL)
				dumpd = pcap_dump_open(handle, unqfnam);

			pcap_loop(handle, (int) STAGE5_CAPTURE_SIZE, callback, NULL); 

                	if (timeout == 0)
                	{
                    		/* cleanup */
                    		alarm(0);
                    		pcap_freecode(&fp);
                    		pcap_close(handle);
                	}
                	else
                    		timeout = 0;

			/* Close dump file. */
			if (dumpd != NULL)
                		pcap_dump_close(dumpd);

			/* Check enough packets received. */
			if (min_threshold_met(5) == 0)
			{
				remove_pcap();
				continue;
			}

			/* Full version of Obeseus contains:
				SYN flood
				RESET flood
				HTTP flood
				UDP flood
				PING flood
				Backscatter packet analysis
				Virtual Darknet/Honeynet subnet analysis
				Snort post processor.
			*/

			/* If only one source address perfom no more checking */
			if (max_num_srcs <= 1)
			{
				calc_bandwidth(NULL, NULL, 0);
				fprintf(data_fptr, "NOT a distributed Obeseus recognised DDOS attack.\n");
			}
			else
			{
				/* Check for a synflood attack. */
				if (is_synflood() == 0)
					calc_bandwidth(NULL, NULL, 0);
			}

			output_seperator();
			
			/* Free the filename string */
			if (unqfnam != NULL)
				free(unqfnam);

			/* Update the blacklist file this iteration. */
			update_blacklist_file();
			fflush(data_fptr);
		} /* End of top pc loop */
		/* Clear up the subnet data for next iteration. */
		memset(top_24_pc, 0, TOP_PC_SIZE);
		top_24_count = 0;
		reset_counts();
		sleep(FIVE_SECS);
	}/* End of while loop */

        /* Close file if opened. */
        if (!no_data_file)
            fclose(data_fptr);

	return 0;
}
